#include <errno.h>
#include <glib/gstdio.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "file.h"

struct _uic_file {
	GIOChannel *ch;
};

uic_file *
uic_file_open(const gchar *path, const gchar *fmt)
{
	GError *err = NULL;
	GIOChannel *ch;
	uic_file *file;

	ch = g_io_channel_new_file(path, fmt, &err);
	if (ch == NULL) {
		switch (err->code) {
		case G_FILE_ERROR_NOENT:
			uic_errnum = UIC_FILE_ERR_NOENT;
			break;
		case G_FILE_ERROR_ISDIR:
			uic_errnum = UIC_FILE_ERR_ISDIR;
			break;
		case G_FILE_ERROR_ACCES:
			uic_errnum = UIC_FILE_ERR_ACCES;
			break;
		case G_FILE_ERROR_EXIST:
			uic_errnum = UIC_FILE_ERR_EXIST;
			break;
		default:
			uic_errnum = UIC_FILE_ERR_OTHER;
			break;
		}
		uic_set_error(err->message);
		g_clear_error(&err);
		return NULL;
	}
	
	if (g_io_channel_set_encoding(ch, NULL, &err) != G_IO_STATUS_NORMAL) {
		uic_set_error(err->message);
		g_clear_error(&err);
		g_io_channel_shutdown(ch, TRUE, NULL);
		g_io_channel_unref(ch);
		return NULL;
	}

	file = g_malloc(sizeof(*file));

	file->ch = ch;

	return file;
}

gint
uic_file_write(uic_file *file, const gchar *buf, gsize cnt)
{
	gsize n;
	GIOStatus st;
	GError *err = NULL;

	if (file->ch->is_writeable == FALSE) {
		uic_set_error("isnt_writable");
		return -1;
	}

	st = g_io_channel_write_chars(file->ch, buf, cnt, &n, &err);

	switch (st) {
	case G_IO_STATUS_NORMAL:
		g_io_channel_flush(file->ch, NULL);
		return n;
		break;
	case G_IO_STATUS_AGAIN:
		uic_errnum = UIC_FILE_ERR_AGAIN;
		uic_set_error("no data");
		break;
	default:
		uic_set_error(err->message);
		g_clear_error(&err);
		break;
	}
	return -1;
}


gint
uic_file_read(uic_file *file, gchar *buf, gsize cnt)
{
	gsize n;
	GIOStatus st;
	GError *err = NULL;

	if (file->ch->is_readable == FALSE) {
		uic_set_error("isnt_readable");
		return -1;
	}

	st = g_io_channel_read_chars(file->ch, buf, cnt, &n, &err);

	switch (st) {
	case G_IO_STATUS_NORMAL:
		return n;
		break;
	case G_IO_STATUS_AGAIN:
		uic_errnum = UIC_FILE_ERR_AGAIN;
		uic_set_error("no data");
		break;
	case G_IO_STATUS_EOF:
		uic_errnum = UIC_FILE_ERR_EOF;
		uic_set_error("end of file");
		return 0;
		break;
	case G_IO_STATUS_ERROR:
	default:
		uic_set_error(err->message);
		g_clear_error(&err);
		break;
	}
	return -1;
}

gint
uic_file_seek(uic_file *file, gint64 offset, GSeekType type)
{
	GIOStatus st;
	GError *err = NULL;

	if (file->ch->is_seekable == FALSE) {
		uic_set_error("isn't seekable");
		return -1;
	}

	st = g_io_channel_seek_position(file->ch, offset, type, &err);
	
	switch (st) {
	case G_IO_STATUS_NORMAL:
		return 0;
	default:
		uic_errnum = UIC_FILE_ERR_OTHER;
		uic_set_error(err->message);
		g_clear_error(&err);
		break;
	}

	return -1;
}


/* Try to create uic_file_st struct.
 * if Fails returns -1 and set uic_errnum
 */

gint
uic_file_stat(const gchar *path, struct uic_file_st *fst)
{

	struct stat st;

	if (g_stat(path, &st) != 0) {
		if (errno == EACCES) {
			uic_errnum = UIC_FILE_ERR_ACCES;
			uic_set_error("access err");
		} else if (errno == ENOENT) {
			uic_errnum = UIC_FILE_ERR_EXIST;
			uic_set_error("no such file or dir");
		} else {
			uic_errnum = UIC_FILE_ERR_OTHER;
			uic_set_error("other err");
		}
		return -1;
	}

	fst->size = (goffset)st.st_size;
	fst->atime = st.st_atime;
	fst->ctime = st.st_ctime;
	fst->mtime = st.st_mtime;

	return 0;
}


gint
uic_file_close(uic_file *file)
{
	GIOStatus st;
	GError *err = NULL;

	if ((st = g_io_channel_shutdown(file->ch, TRUE, &err)) != G_IO_STATUS_NORMAL) {
		uic_set_error(err->message);
		g_clear_error(&err);
		return -1;
	}
	g_io_channel_unref(file->ch);
	g_free(file);

	return 0;
}

